# Script that automatically creates different versions of a routine with different precisions

# Import classes and functions from UtilsRPE
from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.Inserter import replace_real_declaration
from AutoRPE.UtilsRPE.SourceManager import preprocess_sources, parse_sources, load_vault
from AutoRPE.UtilsRPE.Cleaner import remove_comments
from AutoRPE.UtilsRPE.Finder import find_declaration_line
import AutoRPE.UtilsRPE.RegexPattern as re_pattern

from argparse import ArgumentParser
from os import mkdir
from os.path import join, isdir
from shutil import rmtree
import re
import itertools

def get_command_line_arguments():

    help_text = """
    This script takes as arguments the input source (-s/--input-source)
    and the name of a routine (-r/--routine) and creates different versions
    of this routine accouting for all different precision combinations that can happen.
    An additional argument for the vault (-v/--vault) allows to skip the
    creation of the vault if this is provided. Optional output (-o/--output-folder) for creating
    the files there, if not given the outputs are printed out.
    """
    parser = ArgumentParser(usage=help_text)
    parser.add_argument("-v", "--vault", dest="vault", required=False, help="Path to the vault")
    parser.add_argument("-o", "--output-folder", dest="output_folder", required=False, help="Path to outputs folder")
    parser.add_argument("-s", "--input-source", dest="input_source", required=True,
                        help="Path to the processed sources")
    parser.add_argument("-r", "--routine", dest="routine_name", required=True,
                        help="Name of the routine to create different interfaces")
    args = parser.parse_args()

    return args.vault, args.input_source, args.routine_name, args.output_folder


# Few functions
def merge_declaration_lines(code_text):
    lines = [l for l in code_text.split("\n")]
    for _index, _line in enumerate(lines):
        line = remove_comments(_line).strip()
        if line.count("&") and line.count("::"):
            i = 1
            while line and line.strip()[-1] == "&":
                line1 = line
                line2 = remove_comments(lines[_index + i]).strip()
                line = line1.replace("&", "").rstrip() + line2
                i += 1
            line = line.replace("&", "")
            line = re.sub(", +", ", ", line)
            line = re.sub(":  +", ":  ", line)
            lines[_index] = line
            for j in range(1, i):
                lines[_index + j] = ""
    return "\n".join(lines).strip()


def replace_subroutine_name(old_name, new_name, code_text):
    pattern = re_pattern.subroutine_declaration % old_name
    pat = re.compile(pattern, re.I)
    replacement = r"\1 %s" % new_name
    return pat.sub(replacement, code_text)


def obtain_routine_source(routine, sources_path):
    source_file = join(sources_path, "%s.f90" % routine.module.name)
    indices = []
    with open(source_file) as f:
       pattern = re_pattern.subroutine_declaration % routine.name
       lines = [l for l in f]

       for index, line in enumerate(lines):
          line = remove_comments(line).strip()
          if re.search(pattern, line, re.I):
             indices.append(index)
       routine_code = "".join(lines[indices[0]:indices[1]+1])

    routine_code = merge_declaration_lines(routine_code)
    return routine_code, indices


def replace_routine_name(old_name, new_name, code_text):
    pattern = re_pattern.subroutine_declaration % old_name
    pat = re.compile(pattern, re.I)
    replacement = r"\1 %s" % new_name
    return pat.sub(replacement, code_text)

def prepare_vault(vault_path=None):
    path_to_preprocessed_sources = "./tmp_processedsources/"
    # In case no vault has been provided it will create it from the sources in the input sources folder
    if vault_path is None:
        # Paths to source folders
        # Path_to_input_sources
        if not isdir(path_to_preprocessed_sources):
            mkdir(path_to_preprocessed_sources)
        # Format text + replace real declarations with type(rpe_var)
        preprocess_sources(path_to_input_sources, path_to_preprocessed_sources)
        # Obtain source information
        vault = parse_sources(path_to_preprocessed_sources, save_vault=True)
        # Remove generated folder
        rmtree(path_to_preprocessed_sources)
    else:
        vault = load_vault(vault_path)

    return vault


def get_routine(vault, routine_name=None):
    # Find procedure with the routine name
    routine = [p for p in vault.procedures if p.name == routine_name]
    # Check if we found it
    if len(routine) == 0:
        print("Routine %s not found in vault." % routine_name)
        exit(0)

    return routine[0]

def update_line(version_lines, variable, filepath, vault, routines_indices, variable_precision):
    # Find the declaration/original lines
    d_lines = find_declaration_line(variable, filepath, vault)
    # Search in declaration lines
    for d_line in d_lines:
        # Update line if it's between routine lines routine
        if d_line >= routines_indices[0] and d_line <= routines_indices[1]:
            original_line = version_lines[d_line - routines_indices[0]]
            new_line = replace_real_declaration(variable, original_line, variable_precision)
            version_lines[d_line - routines_indices[0]] = new_line

def generate_routine_versions(precision_combinations, sensitive_variables, vault, routine_lines, routine, filepath, indices):
    code_versions = {}
    for c_index, combination in enumerate(precision_combinations):
        new_routine_name = "%s_%04i" % (routine.name, c_index)
        # Search and update real declaration
        for index, variable in enumerate(sensitive_variables):
            update_line(routine_lines, variable, filepath, vault, indices, combination[index])
        # Join the routine lines and uodate nane
        version_text = "\n".join(routine_lines).strip()
        version_text = replace_routine_name(routine.name, new_routine_name, version_text)
        code_versions[new_routine_name] = version_text
    return code_versions

def write_routine_versions(code_versions, output_folder):
    if output_folder is None:
        # If there's no folder, print the outputs to terminal
        for key, code in code_versions.items():
            print(code)
    else:
        # Create folder if it doesn't exists
        if not isdir(output_folder):
            mkdir(output_folder)
        # Create a file for each subroutine interface
        for key, code in code_versions.items():
            with open(output_folder + key, 'w') as f:
                f.write(code)

if __name__ == "__main__":

    # Get command line arguments
    vault_path, path_to_input_sources, routine_name, output_folder = get_command_line_arguments()
    # Get the vault
    vault = prepare_vault(vault_path)
    # Get routine from vault
    routine = get_routine(vault, routine_name)
    # Get the routine code
    routine_code, indices = obtain_routine_source(routine, path_to_input_sources)
    # All possible combinations of precisions x sensitive variables
    sensitive_variables = [v for v in routine.variables if v.is_dummy_argument and v.type == "double"]
    precision_combinations = list(itertools.product(["dp", "sp"], repeat=len(sensitive_variables)))
    # Get filepath
    filepath = join(path_to_input_sources, "%s.f90" % routine.module.name)
    # Generate the different routine versions
    code_versions = generate_routine_versions(precision_combinations, sensitive_variables, vault, routine_code.split("\n"), routine, filepath, indices)

    write_routine_versions(code_versions, output_folder)
